from django.shortcuts import render_to_response
from django.http import HttpResponse, HttpResponseRedirect
from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.template import RequestContext

from project.main.models import Template


#--------------------------------------------------------------------------------------------------

@login_required
def list(request):

    if not request.user.permissions.can_manage_templates:
        return HttpResponseRedirect("/")

    items = Template.objects.all().order_by("order")

    return render_to_response("template/list.html",
        {
            "items": items,
            "active_menu": "templates",
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

@login_required
def edit(request, id = 0):

    if not request.user.permissions.can_manage_templates:
        return HttpResponseRedirect("/")

    id = int(id)
    new_item = id == 0

    if not new_item:
        try:
            item = Template.objects.get(id = id)
        except:
            new_item = True

    if new_item:
        item = Template()

    if request.method == "POST":

        post = lambda key : request.POST.get(key, "")

        if post("operation") == "delete":
            item.delete()
        else:
            if new_item:
                item.author = request.user
                item.order = Template.objects.count()

            item.name = post("name")
            item.description = post("description")
            item.content = post("content")
            item.modified_by = request.user
            item.save()

        return HttpResponseRedirect("/template/list/")

    from project.main.object_factory import ObjectFactory

    return render_to_response("page/edit.html",
        {
            "item": item,
            "object_factory": ObjectFactory(None),
            "is_template": True,
            "active_menu": "templates",
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

@login_required
def delete(request, id):

    if not request.user.permissions.can_manage_templates:
        return HttpResponseRedirect("/")

    try:
        item = Template.objects.get(id = id)
        item.delete()
    except:
        pass

    return HttpResponseRedirect("/template/list/")

#--------------------------------------------------------------------------------------------------

@login_required
def toggle_published(request, id):

    if not request.user.permissions.can_manage_templates:
        return HttpResponseRedirect("/")

    try:
        item = Template.objects.get(id = id)
        item.published = not item.published
        item.save()
    except:
        pass

    return HttpResponseRedirect("/template/list/")

#--------------------------------------------------------------------------------------------------

@login_required
def change_order(request, id, direction):

    if not request.user.permissions.can_manage_templates:
        return HttpResponseRedirect("/")

    try:
        template = Template.objects.get(id = id)
        template.normalize_order()
    except:
        return HttpResponseRedirect("/")

    templates = Template.objects.all()

    if direction == "up":
        for p in templates:
            if p.order == template.order - 1:
                p.order += 1
                template.order -= 1
                p.save()
                template.save()
                break

    elif direction == "dn":
        for p in templates:
            if p.order == template.order + 1:
                p.order -= 1
                template.order += 1
                p.save()
                template.save()
                break

    return HttpResponseRedirect("/template/list/")

#--------------------------------------------------------------------------------------------------
